#!/usr/bin/env python

__version__ = '1.0'
__author__  = "Avinash Kak (kak@purdue.edu)"
__date__    = '2010-August-22'
__url__     = 'http://RVL4.ecn.purdue.edu/~kak/distICP/ICP-1.0.html'
__copyright__ = "(C) 2010 Avinash Kak. Python Software Foundation."

__doc__ = '''

    ICP.py

    Version: ''' + __version__ + '''
   
    Author: Avinash Kak (kak@purdue.edu)

    Date: ''' + __date__ + '''

    INTRODUCTION:

        ICP stands for Iterative Closest Point algorithm.  ICP algorithms
        are used to register two data sets (meaning making one data set 
        spatially congruent with the other data set) by applying iteratively a 
        rotation and translation to one data set until it is congruent with
        the other data set.
    
        In the domain of image processing, ICP can be used to register a data
        image recorded through a sensor against a model image that may be produced
        by a geographic information system (GIS).  A typical application would 
        be a UAV recording images as it flies over a terrain; a successful 
        registration between such sensor images and the the model images produced 
        by an on-board or satellite-connected GIS system would enable precise 
        computation of the position and the orientation of the UAV vis-a-vis 
        the terrain.
    
        The main goal of the pure-Python implementation of ICP presented here 
        is to make it easier to experiment with the different aspects of such 
        algorithms.
    


    USAGE:

        For the case of binary images, a simple usage example would look like:
    
            import ICP
            icp = ICP.ICP( 
                    model_image = "triangle1.jpg",
                    data_image =  "triangle2.jpg",
                    binary_or_color = "binary",
                    iterations = 40,
                    display_step = 1,
                    debug = 1 )
            icp.icp()
            icp.display_results()
    
        For registering a shape in one image with a shape in another image, 
        you'd use the binary case shown above.  By binary image, we mean an image
        in which the only non-zero pixels are on the shape boundaries.
    
        For the case of color images (including the case of grayscale images),
        the simplest calls to the module would look like
    
            import ICP  
            icp = ICP( 
                   model_image = "modelterrain.jpg",
                   data_image  = "image_from_camera.jpg",
                   binary_or_color = "color",
                   iterations = 40,
                   connectivity_threshold = 5,
                   output_image_size = 100,
                   display_step = 1,
                   debug = 1 )
            icp.icp()
            icp.display_results()
    
        where the parameter output_image_size controls the size of the image
        that will actually be used for ICP calculations.  Color (and grayscale)
        images as output by sensors can be very large and it would be impractical
        to process all of the pixel data in the images.  This module first 
        smoothly reduces the size of the images so that the maximum dimension
        does not exceed output_image_size and then carries out ICP processing
        on the reduced-size images.  The parameter connectivity_threshold 
        controls how many of the edge pixels will be chosen for processing.  
        In the current implementation, an edge pixel is chosen if it has at 
        least connectivity_threshold number of other edge pixels in its 5x5 
        neighborhood.
    
        The module also includes a static method gendata() to illustrate how
        one can create simple synthetic images to experiment with this ICP module.
        A call to this method looks like
    
            import ICP
            ICP.gendata()
    
        As currently programmed, the gendata() method constructs images with 
        triangels and straight lines segments.
    


    CONSTRUCTOR PARAMETERS:

        model_image:         The name of the model file   (REQUIRED)
    
        data_image:          The name of the data file    (REQUIRED)
    
        binary_or_color:     Set to 'binary' for binary images and to 'color'
                             for grayscale and color images  (REQUIRED)
    
        output_image_size:   The size to which a large image will be reduced
                             for ICP processing  (DEFAULTS to 100 for large 
                             images)
    
        iterations:          The maximum number of iterations to try (DEFAULTS
                             to 40)
    
        error_threshold:     When the difference in the registration errors 
                             between two successive iterations is less than
                             the error_threshold, we stop iterating.  (DEFAULTS
                             to 0.0)
    
        display_step:        Controls how often the images at the end of successive
                             iterations will be shown in the final display of the
                             results.  For example, when set to 5, every fifth image 
                             will be included for the final display.  (DEFAULTS to 
                             5 when the number of iterations exceed 40 and to 2 
                             otherwise)
    
        connectivity_threshold:  
    
                             This parameter decides as to which edge pixels in a 
                             grayscale or a color image will participate in the ICP 
                             calculations.  For an edge pixel to be accepted, it 
                             must have at least connectivity_threshold number of
                             other edge pixels in a 5x5 neighborhood. (DEFAULTS to 5)
    
        save_output_images:  When set to 1, the module will save the images produced
                             at the end of each iteration in the current directory.
                             Such images are named "__resultX.jpg" for integer values
                             of X that start from 0.  (Defaults to 0)
    
        debug:               When set to 1, the module prints out additional 
                             information useful for debugging purposes.  (DEFAULTS
                             to 0)



    HOW THE RESULTS ARE DISPLAYED:

        The results are displayed using Tkinter graphics (meaning, actually, Tk
        graphics via the Tkinter interface). For binary images, the first row
        shows the model image followed by the data image.  Subsequent rows display
        the data image being translated and rotated as it comes into congruence 
        with the model image.  

        For color (and grayscale) images, the first row displays six images, three
        for the model and three for the data.  In each set of three images, the 
        first image is the actual photograph, the second the edges in the photo, 
        and the third the pixels actually used for ICP calculations.  Subsequent
        rows display the data image being brought into registration with the model
        image through successive iterations.  Note that whereas the ICP calculations
        are carried out on only the sparse set of pixels shown as the third image
        in the model and the data sequences in the first row, the computed 
        transformations are applied to the edge images shown as the second image
        for each.

        The display of results requires access to the "times.ttf" font file for
        the labels that are needed in the display.  The module assumes that
        this file can be located through the paths stored in 'sys.path', or 
        via the paths available through your environment variables, or at the
        following path:

             /usr/share/fonts/truetype/msttcorefonts/times.ttf

        which is typical for a Ubuntu install. If that is not the case, you would 
        need to edit the 11th line of the code shown for display_results() method
        and enter the pathname to where "times.ttf" can be found.


    CAVEATS:

        Within the limitations caused by the fact that ICP calculations can get
        trapped in a local minimum, you can expect to get good results on binary 
        images.  As for color images, what results you get would depend greatly
        on how complex the images are, the quality of edge detection, and how 
        the final selection of pixels for ICP computation is carried out.  The
        implementation shown in this module uses very rudimentary edge 
        detection and even more rudimentary final pixel selection.  (That is 
        because the focus of this module is on ICP and not on digital image
        processing.)  To improve the results on color (and grayscale) images,
        you would need to replace the pixel extraction functions with your own
        implementations.  Another option would be carry out pixel extraction
        separately and to then feed the pixels thus selected as binary images
        to this module.

        When processing color (and grayscale) images, if you are free to 
        decide as to which image to call model and which to call data, choose
        the one that results in fewer pixels for ICP processing as the 
        data image.  You are likely to get better results if you do so.
  
    INSTALLATION:

        The ICP class was packaged using  Distutils.  For installation, execute the 
        following command-line in the source directory (this is the directory that 
        contains the setup.py file after you have downloaded and uncompressed the 
        package):
 
            python setup.py install

        You have to have root privileges for this to work.  On Linux distributions, 
        this will install the module file at a location that looks like

             /usr/lib/python2.6/site-packages/

        If you do not have root access, you have the option of working directly off 
        the directory in which you downloaded the software by simply placing the
        following statements at the top of your scripts that use the ICP class:

            import sys
            sys.path.append( "pathname_to_ICP_directory" )

        To uninstall the module, simply delete the source directory, locate where 
        ICP was installed with "locate ICP" and delete those files.  As  mentioned 
        above, the full pathname to the installed version is likely to look like
        /usr/lib/python2.6/site-packages/ICP*

        If you want to carry out a non-standard install of the ICP module, look up 
        the on-line information on Disutils by pointing your browser to

              http://docs.python.org/dist/dist.html



    THEORETICAL BASIS:   

        The first ICP algorithm was proposed by Paul Besl and Neil McKay in a now 
        celebrated paper that appeared in 1992 in IEEE Transactions on PAMI.  
        Since then various versions of the algorithm have been published by other 
        folks for either speeding up the performance of the algorithm or for
        improving its accuracy.
    
        The algorithm implemented here is as simple as it can be.  We model the 
        relationship between model and data as
    
              R x_d   +   T    =    x_m 
    
        where x_d denote the data points and x_m the model points, each a two-element
        vector.  R is a 2x2 rotation matrix and T a 2-element translation vector.
        Since two planar figures may not be related by the above transformation 
        (even when one figure "appears" to be a rotated version of the other figure) 
        for an arbitrary location of the origin, move the origin to the mean point
        of the model by
    
              x_m   =   x_m  -   mean( x_m )
                                                           
              x_d   =   x_d  -   mean( x_m )
    
        With regard to the calculation of R and T, let's express the data points and 
        the CORRESPONDING model points as
    
             [x_d1, x_d2, ....., x_dn]  
    
        and their CORRESPONDING model points
    
             [x_m1, x_m2, ....., x_mn]
    
        We can now write the following for estimating the rotation matrix:
    
             R . A   =  B
    
        where
    
             A  =  [x_d1, x_d2, ....., x_dn]  
    
             B  =  [x_m1 - T, x_m2 - T, ....., x_mn - T]
    
        Both A and B are 2xn matrices.  We can now write
    
             R . A . A^t =  B . A^t
    
        where A^t is the transpose of A.  So, we have the following as a
        least mean squares estimate for R:
    
             R    =    B . A^t . ( A . A^t )^-1
    
        Since such an R may not obey the strict properties that must apply
        to a rotation matrix (it must be orthonormal), we condition it by
        first subjecting it to a singular value decomposition:
    
            U.S.V^t  =  svd( R )
    
        and then writing the following for a better estimate for R:
     
            R    =   U . V^t
    
        We can now estimate T by
    
            T    =  mean( x_m )   -   mean( R x_d )        
    
        The above assumes that we are carrying out an one-shot calculation of R 
        and T.  But ICP is iterative.  The following applies to an iterative 
        implementation of the above:
    
        For an iterative assessment of R and T, let's assume that we can decompose 
        R as
    
            R = \deltaR . R_0
    
        where we have previously calculated R_0 and now we wish to refine the 
        estimate with the calculation of \deltaR.  Plugging the above in the previous 
        formulation:
    
            \deltaR . R_0 . A . A^t =  B . A^t
    
        implying
    
            \deltaR . R_0   =  B . A^t . ( A . A^t )^-1
    
        which is to say:
    
            \deltaR   =  B . A^t . ( A . A^t )^-1 .  R_0^t
    
        After you have calculated \deltaR, you can update the rotation matrix by
    
            R = \deltaR . R_0
    
        At the end of each such update of the rotation matrix, the calculation of 
        the translation vector remains the same as before
    
            T    =  mean( x_m )   -   mean( R .  x_d )               
    
        We can therefore write down the following steps for ICP computation:
    
        STEP 1:
                   x_m   =   x_m  -  mean(x_m)
                   x_d   =   x_d  -  mean(x_m)
    
    
        STEP 2:
                   Initialize R and T:
    
                   R  =    1  0
                           0  1
    
                   T  =    0
                           0
    
                   old_error = inf
    
    
        STEP 3:     error = (1/N) \sum dist( x_m - (R * x_d + T) )
    
    
        STEP 4:     diff_error = abs(old_error - error)
                    if (diff_error > threshold):
                        old_error = error
                    else:
                        break
    
        STEP 5:    for each x_d find its closest x_m by finding that x_m
                   which minimizes the squared difference between the
                   two sides of
    
                          R . x_d  +   T  =   x_m
    
        STEP 6:     A  =  [x_d1, x_d2, ....., x_dn]  
                    B  =  [x_m1 - T, x_m2 - T, ....., x_mn - T]
    
                    Note that A will remain the same for ICP iterations,
                    but B can change with each iteration depending on the
                    what corresponding pixels are found for each data pixel.
    
        STEP 7:     AATI =  A^t inverse(A * A^t)
    
        STEP 8:     R_update =  B * AATI * R.T
    
        STEP 9:     U,S,VT =  svd(R_update)
                    deter =   determinant(U * VT)
                    U[0,1] = U[0,1] * determinant
                    U[1,1] = U[1,1] * determinant
    
                    R_update = U * VT
    
        STEP 10:    R =  R_update * R
    
        STEP 11:    T  =  mean(x_m) - mean( R * x_d )
    
        STEP 12:    Back to Step 3
    

    ACKNOWLEDGMENTS:  
 
        The author would like to thank Chad Aeschliman and Kyuseo Han for 
        their considerable help with the debugging of the code presented here.


    ABOUT THE AUTHOR:  

        Avi Kak is in the last stages of finishing up his multi-year Objects
        Trilogy project.  See his web page at Purdue for what this project is
        all about.
'''

import Image
import ImageFilter
import ImageFont
import ImageDraw
import scipy
import scipy.linalg
import ImageTk
import Tkinter
import sys, os.path

#____________________  Support functions _______________________

# calculate Euclidean distance between two points p and q and
# return the distance as a scalar:
def _euclidean(p, q):
    p,q = scipy.array(p), scipy.array(q)
    return scipy.sqrt( scipy.dot(p-q,p-q) )

# calculate Euclidean distance from a point p to the list of
# points in data and return a list of all such distances
def _dist(p, data):
    return [_euclidean(p,q) for q in data]

def _indexAndLeastDistance(distList):
    minVal = min(distList)
    return distList.index(minVal), minVal    

def _difference(p,q):
    return p[0]-q[0],p[1]-q[1]

def _mean_coordinates(coords_list):
    '''
    Returns a pair of values in the form of a tuple, the pair
    consisting of the mean values for the x and the y coordinates.
    '''
    mean = reduce(lambda x,y: x+y, [x[0] for x in coords_list]), \
           reduce(lambda x,y: x+y, [x[1] for x in coords_list])
    mean = mean[0]/float(len(coords_list)), mean[1]/float(len(coords_list))
    return mean

#---------------------- ICP Class Definition ---------------------

class ICP(object):

    def __init__(self, *args, **kwargs ):
        if args:
            raise ValueError(  
                   '''ICP constructor can only be called
                      with keyword arguments for the following
                      keywords: model_image, data_image, binary_or_color,
                      output_image_size, iterations, error_threshold, 
                      display_step,connectivity_threshold, or
                      save_output_images''')       

        model_image = data_image = output_image_size = iterations = None
        error_threshold = display_step = connectivity_threshold = None
        debug = save_output_images = None

        if kwargs.has_key('model_image'): model_image=kwargs.pop('model_image')
        if kwargs.has_key('data_image'):  data_image=kwargs.pop('data_image')
        if kwargs.has_key('binary_or_color'): \
                                binary_or_color=kwargs.pop('binary_or_color')
        if kwargs.has_key('output_image_size'):  \
                            output_image_size=kwargs.pop('output_image_size')
        if kwargs.has_key('iterations'):  iterations=kwargs.pop('iterations')
        if kwargs.has_key('error_threshold'):  \
                            error_threshold=kwargs.pop('error_threshold')
        if kwargs.has_key('display_step'):  display_step=kwargs.pop('display_step')
        if kwargs.has_key('connectivity_threshold'):  \
                  connectivity_threshold=kwargs.pop('connectivity_threshold')
        if kwargs.has_key('debug'): debug=kwargs.pop('debug')
        if kwargs.has_key('save_output_images'): \
                             save_output_images=kwargs.pop('save_output_images')
        if len(kwargs) != 0:
            raise ValueError('''You have provided unrecognizable keyword args''')

        if model_image: 
            self.model_im = Image.open(model_image)
        else:
            raise ValueError('''You must specify a model image''')
        if data_image: 
            self.data_im =  Image.open(data_image)
        else:
            raise ValueError('''You must specify a data image''')
        if binary_or_color:
            self.binary_or_color = binary_or_color
        else:
            raise ValueError('''You must specify either "binary" or "color" ''')
        if output_image_size:
            self.output_image_size = output_image_size
        elif self.model_im.size[0] <= 100:
                self.output_image_size = self.model_im.size[0]
        else:
            self.output_image_size = 100
        if iterations:
            self.iterations = iterations
        else:
            self.iterations = 10
        if connectivity_threshold:
            self.connectivity_threshold = connectivity_threshold
        else:
            self.connectivity_threshold = 5
        if error_threshold:
            self.error_threshold = error_threshold
        else:
            self.error_threshold = 0.0
        if display_step:
            self.display_step = display_step
        else:
            if self.iterations > 40:
                self.display_step = 5
            else:
                self.display_step = 2
        if debug:
            self.debug = debug
        else:
            self.debug = 0
        if save_output_images:
            self.save_output_images = save_output_images
        else:
            self.save_output_images = 0
        self.tk_images = []

    def extract_model_pixels(self):
        width, height =  self.model_im.size
        if self.debug: print "model image of width ", width, " and height ", height
        if self.binary_or_color == 'color':
            self.model_im.thumbnail( (self.output_image_size,\
                                      self.output_image_size), Image.ANTIALIAS )
            self.model_edge_im = \
                self.model_im.filter(ImageFilter.FIND_EDGES).convert('L').convert('1')
            image_processing_results = \
                     self.extract_pixels_from_color_image( self.model_edge_im )
            self.model_list = image_processing_results[0]
            self.sparse_model_list = image_processing_results[1]
            if self.debug: 
                print "sparse_model_list: ", self.sparse_model_list
                print "number of pixels selected for ICP: ", len(self.sparse_model_list)
                print "number of all edge pixels: ", len(self.model_list)
            self.sparse_model_im = image_processing_results[2]
        else:
            self.model_im = self.model_im.convert("1")
            for i in range(3,width-3): 
                for j in range(3,height-3):
                    if ( self.model_im.getpixel((i,j)) != 0 ):  
                        self.model_list.append( (i,j) )       
        if self.debug:
            print "model list", self.model_list
            print "Number of pixels in model_list: ", len(self.model_list)
            print "\n\n"


    def extract_data_pixels(self):
        '''
        Since extract_model_pixels() and extract_data_pixels() have the same
        logic, why can't we have just one method called with two different
        arguments?  These exist as separate methods to allow for the future
        possibility that one may want to process the model and the data images
        very differently.  Presumably, the model images will be noiseless and
        will be output by some GIS system.  On the other hand, the data images
        can be expected to be noisy and may suffer from optical and other 
        distortions.
        '''
        if self.binary_or_color == 'color':
            self.data_im.thumbnail( (self.output_image_size,\
                                     self.output_image_size), Image.ANTIALIAS )
            self.data_edge_im = \
                self.data_im.filter(ImageFilter.FIND_EDGES).convert('L').convert('1')
            image_processing_results = \
                  self.extract_pixels_from_color_image( self.data_edge_im )
            self.data_list = image_processing_results[0]
            self.sparse_data_list = image_processing_results[1]
            if self.debug: 
                print "sparse_data_list: ", self.sparse_data_list
                print "number of data pixels selected for ICP: ", \
                                             len(self.sparse_data_list)
                print "number of all model edge pixels: ", len(self.data_list)
            self.sparse_data_im = image_processing_results[2]
        else:
            self.data_im = self.data_im.convert("1")
            (width,height) = self.data_im.size
            for i in range(width): 
                for j in range(height):
                    if self.data_im.getpixel((i,j)) != 0:  
                        self.data_list.append( (i,j) )       
        if self.debug:
            print "data list", self.data_list
            print "Number of pixels in data_list: ", len(self.data_list)
            print "\n\n"


    def extract_pixels_from_color_image(self, image):
        '''
        This very simple routine would need to be either replaced in this
        class or overridden in a subclass of ICP for a more practical 
        approach to the selection of pixels for ICP calculation.  At this 
        time, all we do is to choose a pixel if it has at least 
        connectivity_threshold number of pixels in its 5x5 neighborhood.  
        This is done with the hope that pixels on edges are likely 
        to be selected with this criterion.
        '''
        (width,height) = image.size
        im_sparse =  Image.new("1", (width,height), 0)
        sparse_edge_pixels_list = []   
        all_edge_pixels_list = []                
        for i in range(3,width-3):
            for j in range(3,height-3):
                count = 0
                if ( image.getpixel((i,j)) != 0 ):
                    all_edge_pixels_list.append( (i,j) )
                    for k in (-2,-1,1,2):
                        for l in (-2,-1,1,2):
                            if ( image.getpixel((i+k,j+l)) == 255 ):
                                count = count + 1;
                if (count >= self.connectivity_threshold): 
                    im_sparse.putpixel((i,j),255)
                    sparse_edge_pixels_list.append( (i,j) )
        return all_edge_pixels_list, sparse_edge_pixels_list, im_sparse

    def move_to_model_origin(self):
        '''
        Since two patterns in a plane, even when one appears to be a
        rotated version of the other, may be not be related by a 
        Euclidean transform for an arbitrary placement of the origin,
        we will assume that the origin for ICP calculations will be 
        at the "center" of the model image.  Now our goal becomes to
        find an R and a T that will make the data pattern congruent 
        with the model pattern.
        '''
        if self.binary_or_color == 'color':
            self.model_mean =  \
                (scipy.matrix(list(_mean_coordinates(self.sparse_model_list)))).T
            self.zero_mean_sparse_model_list = [(p[0] - self.model_mean[0,0], \
                         p[1] - self.model_mean[1,0]) for p in self.sparse_model_list]
            self.zero_mean_model_list = [(p[0] - self.model_mean[0,0], \
                         p[1] - self.model_mean[1,0]) for p in self.model_list]
            self.zero_mean_sparse_data_list = [(p[0] - self.model_mean[0,0], \
                            p[1] - self.model_mean[1,0]) for p in self.sparse_data_list]
            self.zero_mean_data_list = [(p[0] - self.model_mean[0,0], \
                                   p[1] - self.model_mean[1,0]) for p in self.data_list]
        else: 
            # Need to do the following separately because for we calculate 
            # model mean from only the sparse set for color images
            self.model_mean = (scipy.matrix(list(_mean_coordinates(self.model_list)))).T
            self.zero_mean_model_list = [(p[0] - self.model_mean[0,0], \
                                p[1] - self.model_mean[1,0]) for p in self.model_list]
            self.zero_mean_data_list = [(p[0] - self.model_mean[0,0], \
                                p[1] - self.model_mean[1,0]) for p in self.data_list]
            if self.debug:
                print "model mean:\n", self.model_mean
                print "zero mean model list: ", self.zero_mean_model_list
                print "zero mean data list: ", self.zero_mean_data_list
                print "\n\n"

    def construct_A_matrix(self):
        '''
        In the plane whose origin at the model center, the relationship
        between the model points x_m and the data points x_d is given by

             R . x_d  +  T  =  x_m
        Let the list of the n chosen data points be given by
             A =  [x_d1, x_d2, .....,   x_dn]
        We can now express the relationship between the data and the 
        model points by
             R . A  = B
        where B is the list of the CORRESPONDING model points after 
        we subtract the translation form each:
             B =  [x_m1 - T, x_m2 - T, ......,  x_mn - T] 
        Eventually our goal will be to estimate R from the R.A = B 
        relationship.
        '''
        if self.binary_or_color == 'color':
            data = self.zero_mean_sparse_data_list
        else:
            data = self.zero_mean_data_list
        self.A = scipy.matrix( [ [p[0] for p in data], [p[1] for p in data] ] )

    def construct_AATI_matrix(self):
        '''
        So we want to estimate R from R.A = B.  If we had to construct a 
        one-shot estimate for R (note ICP is iterative and not one-shot),
        we would write R.A=B as 
                 R.A.A^t  =  B.A^t
        and then
                 R =  B . A^t . (A . A^t)^-1
        We will group together what comes after B on the right hand side
        and write
                 AATI  =  A^t . (A . A^t)^-1
        In the ICP implementation, this matrix will remain the same for all
        the iterations.
        '''
        A = self.A
        self.AATI = A.T * scipy.linalg.inv( A * A.T )

    def initialize(self):
        self.R = scipy.matrix( [[1.0, 0.0],[0.0, 1.0]] )
        self.T = (scipy.matrix([[0.0, 0.0]])).T
        self.model_list = []
        self.data_list = []
        self.sparse_model_list = []
        self.sparse_data_list = []
        self.zero_mean_sparse_model_list = []
        self.zero_mean_sparse_data_list = []
        self.zero_mean_model_list = []
        self.zero_mean_data_list = []

    def setSizeForDisplay(self):
        if self.binary_or_color == 'color':
            self.displayWidth, self.displayHeight = self.data_edge_im.size
        else:
            self.displayWidth, self.displayHeight = self.model_im.size

    def icp(self):
        self.initialize()
        self.extract_model_pixels()
        self.extract_data_pixels()
        self.move_to_model_origin()
        self.construct_A_matrix()
        self.construct_AATI_matrix()
        self.setSizeForDisplay()
        old_error = float('inf')
        iteration = 0
        R,T = self.R, self.T
        if self.binary_or_color == "color":
            model = self.zero_mean_sparse_model_list
            data = self.zero_mean_sparse_data_list
        else:
            model = self.zero_mean_model_list
            data = self.zero_mean_data_list
        while 1:
            print "\n>>>>>>>>>>   STARTING ITERATION ", iteration, " OUT OF ", self.iterations, "\n"
            if iteration == self.iterations: break

            data_matrix = [ R * (scipy.matrix( list(data[p]) )).T + T \
                                                for p in range(len(data)) ] 
            data_transformed = [ ( p[0,0], p[1,0] ) for p in data_matrix]

            # For every data point find the closest model point.  The set of such
            # model points will be called the matched_model set of points
            # The following returns a list of pairs, the first the index of 
            # the model point that was found closest to the data point in question,
            # and the second the actual minimal distance
            leastDistMapping = [_indexAndLeastDistance(_dist(p,model)) \
                                                for p in data_transformed]
            matched_model = [model[p[0]] for p in leastDistMapping]
            matched_model_mean = scipy.matrix(list(_mean_coordinates(matched_model))).T
            error = reduce(lambda x, y: x + y, [x[1] for x in leastDistMapping])
            error = error / len(leastDistMapping)
            print "old_error: ", old_error, "    error: ", error
            print "\n"
            diff_error = abs(old_error - error)
            if (diff_error > self.error_threshold):
                old_error = error
            else:
                self.iterations = iteration
                break
            AATI = self.AATI
            B = scipy.matrix([ [p[0] - T[0,0] for p in matched_model], \
                               [p[1] - T[1,0] for p in matched_model] ])
            R_update = B * AATI * R.T
            [U,S,VT] = scipy.linalg.svd(R_update)
            U,VT = scipy.matrix(U), scipy.matrix(VT) 
            deter = scipy.linalg.det(U * VT)
            U[0,1] = U[0,1] * deter
            U[1,1] = U[1,1] * deter
            R_update = U * VT
            R = R_update * R
            print "Rotation:\n", R
            print "\n"
            # Rotate the data for estimating the translation T
            data_matrix2 = [ R * (scipy.matrix( list(data[p]))).T  \
                                                for p in range(len(data)) ] 
            data_transformed2 = [ ( p[0,0], p[1,0] ) for p in data_matrix2]
            data_transformed_mean = \
                  scipy.matrix(list(_mean_coordinates(data_transformed2))).T
            T = matched_model_mean - data_transformed_mean  
            print "Translation:\n", T
            print "\n"
            # Now apply the R,T transformation to the original data for updated image
            # representation of the result at the end of this iteration
            if self.binary_or_color == 'color':
                data_matrix_new = \
                 [ R * (scipy.matrix( list(self.zero_mean_data_list[p]))).T + T \
                                        for p in range(len(self.zero_mean_data_list)) ] 
            else:
                data_matrix_new = [ R * (scipy.matrix( list(data[p]))).T + T \
                                              for p in range(len(data)) ] 
            data_transformed_new = \
                [ ( p[0,0] + self.model_mean[0,0], p[1,0] + self.model_mean[1,0] ) \
                                                       for p in data_matrix_new]
            result_im = Image.new("1", (self.displayWidth,self.displayHeight), 0)
            for p in data_transformed_new:
                x,y = int(p[0]), int(p[1])
                if ( (0 <= x < self.displayWidth) and (0 <= y < self.displayHeight ) ):
                    result_im.putpixel( (x,y), 255 )
            result_im.save( "__result" + str(iteration) + ".jpg")
            iteration = iteration + 1
        self.R,self.T = R,T


    def display_results(self):
        rootWindow = Tkinter.Tk()
        rootWindow.geometry("1200x750+50+50") 
        cellwidth = self.output_image_size
        padding = 10
        padded_cellwidth = cellwidth + 2*padding
        if cellwidth > 80:
            fontsize = 20
        else:
            fontsize = 10
        import os.path
        if os.path.isfile("times.ttf"):
            font = ImageFont.truetype("times.ttf", fontsize)
        elif os.path.exists("/usr/share/fonts/truetype/msttcorefonts/times.ttf"):
            font = ImageFont.truetype( \
                "/usr/share/fonts/truetype/msttcorefonts/times.ttf", fontsize)
        else:
            print "Unable to find the font file 'times.ttf' needed for displaying the results"
            sys.exit(1)
        textImage1 = Image.new( "F", (self.displayWidth,self.displayHeight), 200 )
        draw = ImageDraw.Draw(textImage1)
        draw.text((10,10), "Model:", font=font) 
        textImage2 = Image.new( "F", (self.displayWidth,self.displayHeight), 200 )
        draw = ImageDraw.Draw(textImage2)
        draw.text((10,10), "Data:", font=font) 
        textImage3 = Image.new( "F", (self.displayWidth,self.displayHeight), 200 )
        draw = ImageDraw.Draw(textImage3)
        draw.text((10,10), "Results:", font=font) 
        if self.debug: print "cell width for display: ", cellwidth, "\n\n"
        if self.binary_or_color == 'color':
            self.tk_images.append(ImageTk.PhotoImage( textImage1 ))
            self.tk_images.append(ImageTk.PhotoImage( self.model_im ))
            self.tk_images.append(ImageTk.PhotoImage( self.model_edge_im ))            
            self.tk_images.append(ImageTk.PhotoImage( self.sparse_model_im ))
            self.tk_images.append(ImageTk.PhotoImage( textImage2 ))
            self.tk_images.append(ImageTk.PhotoImage( self.data_im ))
            self.tk_images.append(ImageTk.PhotoImage( self.data_edge_im ))
            self.tk_images.append(ImageTk.PhotoImage( self.sparse_data_im ))
        else:
            self.tk_images.append(ImageTk.PhotoImage( textImage1 ))
            self.tk_images.append(ImageTk.PhotoImage( self.model_im ))
            self.tk_images.append(ImageTk.PhotoImage( textImage2 ))
            self.tk_images.append(ImageTk.PhotoImage( self.data_im ))
        for i in range(len(self.tk_images)):
            Tkinter.Label(rootWindow,image=self.tk_images[i],\
                    width=cellwidth).grid(row=0,column=i,padx=10,pady=10)
        print "Will display ", self.iterations, "result images\n\n"
        resultImageLabel = ImageTk.PhotoImage(textImage3)
        Tkinter.Label(rootWindow,image=resultImageLabel,\
                    width=cellwidth).grid(row=1,column=0,padx=10,pady=10)
        tkim = [None] * self.iterations
        for i in range(0,self.iterations,self.display_step):
            tkim[i] = ImageTk.PhotoImage( Image.open( "__result" + str(i) + ".jpg" ) )
            j = i / self.display_step
            if ((j+1)*padded_cellwidth < 1200):
                Tkinter.Label(rootWindow,image=tkim[i],\
                    width=cellwidth).grid(row=1,column=j+1,padx=10,pady=10)
            elif ((j+1)*padded_cellwidth < 2400):
                j = j - (1200-padded_cellwidth) / padded_cellwidth
                Tkinter.Label(rootWindow,image=tkim[i],\
                    width=cellwidth).grid(row=2,column=j,padx=10,pady=10)
            elif ((j+1)*padded_cellwidth < 3600):
                j = j - (2400-padded_cellwidth) / padded_cellwidth
                Tkinter.Label(rootWindow,image=tkim[i],\
                    width=cellwidth).grid(row=3,column=j,padx=10,pady=10)
            else: 
                j = j - (3600-padded_cellwidth) / padded_cellwidth
                Tkinter.Label(rootWindow,image=tkim[i],\
                    width=cellwidth).grid(row=4,column=j,padx=10,pady=10)
        Tkinter.mainloop()

    def __del__(self):
        if not self.save_output_images:
            import glob,os
            for filename in glob.glob( '__result*' ): os.unlink(filename)

    @staticmethod
    def gendata():
        '''
        The code here is just the simplest example of synthetic data
        generation for experimenting with ICP.  You can obviously 
        construct more complex model and data images by calling on the
        other shape drawing primitives of the ImageDraw class.  When
        specifying coordinates, note the following

               .----------> positive x
               |
               |
               |        
               V
             positive y

        A line is drawn from the first pair (x,y) coordinates to the
        second pair.
        '''
        s = 100
        data = Image.new( "L", (s,s), 0 )
        draw = ImageDraw.Draw(data)
        draw.line( (0.8*s, 0, 0, s/2), fill=255 )
        draw.line( (0.8*s, 0, s-1, 0.2*s), fill=255 )
        draw.line( (s-1, 0.2*s, 0, s/2), fill=255 )
        del draw
        data.save( "triangleA.jpg" )

        data = Image.new( "L", (s,s), 0 )
        draw = ImageDraw.Draw(data)
        draw.line( (0, s/2, s-1, 0.8*s), fill=255 )
        draw.line( (0, s/2, 0.8*s, s-1), fill=255 )
        draw.line( (0.8*s, s-1, s-1, 0.8*s), fill=255 )
        del draw
        data.save( "triangleB.jpg" )

        theta = 20
        radians = theta * math.pi / 180
        offset = s * scipy.tan(radians)
        print "offset: ", offset
        data = Image.new( "L", (s,s), 0 )
        draw = ImageDraw.Draw(data)
        draw.line( (0,(s/2 + offset/2)) + (s-1, (s/2 - offset/2)), fill=255 )
        del draw
        data.show()
        data.save( "linedataA.jpg" )
        
        theta = 70
        radians = theta * math.pi / 180
        offset = s * scipy.tan(radians)
        print "offset: ", offset
        data = Image.new( "L", (s,s), 0 )
        draw = ImageDraw.Draw(data)
        draw.line( (0,(s/2 + offset/2)) + (s-1, (s/2 - offset/2)), fill=255 )
        del draw
        data.show()
        data.save( "linedataB.jpg" )
        
#------------------------- End of ICP Class Definition  ---------------------------

#-------------------------    Test code follows         ---------------------------

if __name__ == '__main__': 

    #ICP.gendata()

    '''
    icp = ICP( 
               model_image = "triangle1.jpg",
               #model_image = "linemodel.jpg", 
               data_image =  "triangle2.jpg",
               #data_image =  "linedata.jpg",
               binary_or_color = "binary",
               iterations = 40,
               display_step = 1,
               debug = 1 )
    icp.icp()
    icp.display_results()
    '''

    icp = ICP( 
               model_image = "Examples/book1.jpg",
               data_image = "Examples/book2.jpg",
               binary_or_color = "color",
               iterations = 40,
               connectivity_threshold = 5,
               output_image_size = 100,
               display_step = 1,
               debug = 0 )

    icp.icp()
    icp.display_results()
